// Agent monitor in project infinite2.mas2j

/* Initial beliefs and rules */

timeout(4000). 

/****************************************************************/
/********** 				PROTOCOLS					*********/
/****************************************************************/

global_type(abp_no_constr, S0) :- // Ancona, Mascardi, Barbieri FI 2013 
	Msg1 = sa(mess(1), 0) &
	Msg2 = sa(mess(2), 0) &
	Ack1 = sa(ack(1), 0) &
	Ack2 = sa(ack(2), 0) &
	S0 = seq(Msg1, S1) & 
	S1 = choice([ seq(Ack1, S3), seq(Msg2, S2) ]) &
	S2 = choice([ seq(Ack1, S4), seq(Ack2, seq(Ack1,S0))]) &
	S3 = seq(Msg2, S4) &
	S4 = choice([ seq(Ack2, S0), seq(Msg1, S6) ]) &
	S6 = choice([ seq(Ack2, S1), seq(Ack1, seq(Ack2,S3))]).

/****************************************************************/
/********** 		SENDING ACTION TYPES				*********/
/****************************************************************/

has_type(msg(alice_abp2_no_constr, bob_abp2_no_constr, tell, mess(N)), mess(N)).
has_type(msg(bob_abp2_no_constr, alice_abp2_no_constr, tell, ack(N)), ack(N)).
   

/****************************************************************/
/********** 				NEXT PREDICATE				*********/
/****************************************************************/

next(0,seq(sa(AType,N),T),AMessage,T,N) :- has_type(AMessage, AType).
next(M,seq(sa(AType),T1),AMessage,T1,N) :- has_type(AMessage, AType) & M > 0 & N = M - 1.
next(M,choice([T1,_]),A,T2,N) :- next(M,T1,A,T2,N). 
next(M,choice([_,T1]),A,T2,N) :- next(M,T1,A,T2,N).
next(N,fork(T1,T2),A,fork(T3,T2),M) :- next(N,T1,A,T3,M). 
next(N,fork(T1,T2),A,fork(T1,T3),M) :- next(N,T2,A,T3,M).
next(N,fork(T1,T2),A,fork(T3,T4),P) :- next(N,T1,A,T3,M) & M > 0 & next(M,T2,A,T4,P). 
next(N,fork(T1,T2),A,fork(T4,T3),P) :- next(N,T2,A,T3,M) & M > 0 & next(M,T1,A,T4,P). 
next(M,cat(T1,T2),A,cat(T3,T2),N) :- next(M,T1,A,T3,N).
next(M,cat(T1,T2),A,T3,N) :- empty(T1) & next(M,T2,A,T3,N). 


/****************************************************************/
/********** 				EMPTY PREDICATE				*********/
/****************************************************************/

empty(lambda).
empty(choice([T1,_])) :- empty(T1).
empty(choice([_,T1])) :- empty(T1).
empty(fork(T1,T2)) :- empty(T1) & empty(T2).
empty(cat(T1,T2)) :- empty(T1) & empty(T2).

/****************************************************************/
/********** 			TYPE_CHECK PREDICATE			*********/
/****************************************************************/

type_check(msg(S, R, P, C), NewState) :- 
   .print("\nFrom type_check 1, looking for the last state ") &  current_state(LastState) & .wait(5000) & .print("\nFrom type_check 2, the last state is ", LastState) & .wait(5000) &
   .print( "\n", msg(S, R, P, C), "\n") & 
   .print("inside type_check: LastState = ", LastState , "\n") & 
   .wait(4000).
   //&
   //next(0, LastState, msg(S, R, P, C), NewState, 0).
   

/****************************************************************/
/********** 				CURRENT STATE				*********/
/****************************************************************/   
   
current_state(initial).

/****************************************************************/
/********** 				INITIAL GOAL				*********/
/****************************************************************/

!test(global_type(abp_no_constr)).

/* Plans */

/* Monitor's plans */

@test[atomic]
+!test(global_type(Name)) : current_state(initial) & global_type(Name, G) 
   <-  .print("\n\nFrom test 1, moving to ", G); !move_to_state(G).

@move2state[atomic]
+!move_to_state(NewState) : current_state(LastState) 
   <- .wait(3000); .print("\n\nFrom move_to_state 1, the old state is ", LastState); -current_state(_); .wait(3000); .print("\n\nFrom move_to_state 2, the old state is ", LastState);
      +current_state(NewState); .wait(3000); .print("\n\nFrom move_to_state 3, the new state is ", NewState, "\n\nNow a unification error will take place... better to pause the execution...\n"); .wait(7000).	

@successfulMove[atomic]
+!type_check_message(msg(S, R, P, C)) : current_state(Current) & type_check(msg(S, R, P, C), NewState)  
   <- !move_to_state(NewState);
      .print("\nMessage ", msg(S, R, P, C) , "\nleads from state ", Current, " \nto state ", NewState , "\n");
	 // .wait(500);
      .send(S, tell, ok_check(msg(S, R, P, C))).

@failingMoveAndProtocol
+!type_check_message(msg(S, R, P, C)) : current_state(Current) 
                                        & global_type(Name, _)
   <- .print("\n*** DYNAMIC TYPE-CHECKING ERROR ***\nMessage ", msg(S, R, P, C), " received within protocol ", Name, "\ncannot be accepted in the current state ", Current, "\n");
       !move_to_state(initial).

@messageReceptionKO
+msg(S, R, P, C): current_state(initial)
   <- .print("\n*** DYNAMIC TYPE-CHECKING ERROR ***\nMessage ", msg(S, R, P, C), " received but no protocol is ongoing successfully\n").
   
@messageReceptionOK
+msg(S, R, P, C)[source(S)]: timeout(MS)
   <- -msg(S, R, P, C)[source(S)]; 
      !type_check_message(msg(S, R, P, C));
	  !check_progress.
	  
+!check_progress : timeout(MS) 
   <- .wait({+msg(S1, R1, P1, C1)}, MS, Event); 
	  !aux_check_progress(Event).
	  
+!aux_check_progress(Event) : timeout(MS) 
                                           & Event < MS. 

+!aux_check_progress(Event) : timeout(MS) 
                                           & current_state(Current) 
										   & Event >= MS  
   <- .print("\n*** WARNING ***\nNo progress for ", Event, " milliseconds in the current state ", Current, "\n"); 
   !check_progress. 
