// Agent monitor in project infinite2.mas2j

/* Initial beliefs and rules */

timeout(4000). 

/****************************************************************/
/********** 				PROTOCOLS					*********/
/****************************************************************/

/* Brokering protocol, involves b, c, s */
global_type(brokering, Glob) :- // Esop 2012 paper by Pierre-Malo Denielou and Nobuko Yoshida, page 2
   Merge = choice([Off,Fork]) &
   Off=
   seq(sa(offer, 0),
      seq(sa(counter, 0), Merge)) &
   Fork= fork(seq(sa(final, 0),lambda),
              seq(sa(result, 0),lambda)) &
   Glob = seq(sa(item, 0),Merge). 
   
/****************************************************************/
/********** 		SENDING ACTION TYPES				*********/
/****************************************************************/

/* Types for the brokering protocol */	  
has_type(msg(B, C, tell, m(offer(Item, Offer), _CId)), offer) :- role(B, broker) & role(C, client) & string(Item) & int(Offer).
has_type(msg(C, B, tell, m(counter(Item, Offer), _CId)), counter) :- role(C, client) & role(B, broker) & string(Item) & int(Offer).
has_type(msg(B, S, tell, m(final(Res, Client, Item, Offer), _CId)), final) :- role(B, broker) & role(S, seller) & string(Res) & string(Client) & string(Item) & int(Offer).
has_type(msg(B, C, tell, m(result(Res, Item, Offer), _CId)), result) :- role(B, broker) & role(C, client) & string(Res) & string(Item) & int(Offer).
has_type(msg(S, B, tell, m(item(Client, Item), _CId)), item) :- role(S, seller) & role(B, broker) & string(Client) & string(Item).

string(_).
int(_).
role(b_conv, broker).   
role(b_conv2, broker).
role(c_conv, client).
role(c_conv2, client).
role(s_conv, seller).
role(s_conv2, seller).
/****************************************************************/
/********** 				NEXT PREDICATE				*********/
/****************************************************************/

next(0,seq(sa(AType,N),T),AMessage,T,N) :- has_type(AMessage, AType).
next(M,seq(A,T1),A,T1,N) :- M > 0 & N = M - 1.
next(M,choice([T1,_]),A,T2,N) :- next(M,T1,A,T2,N). 
next(M,choice([_,T1]),A,T2,N) :- next(M,T1,A,T2,N).
next(N,fork(T1,T2),A,fork(T3,T2),M) :- next(N,T1,A,T3,M). 
next(N,fork(T1,T2),A,fork(T1,T3),M) :- next(N,T2,A,T3,M).
next(N,fork(T1,T2),A,fork(T3,T4),P) :- next(N,T1,A,T3,M) & M > 0 & next(M,T2,A,T4,P). 
next(N,fork(T1,T2),A,fork(T4,T3),P) :- next(N,T2,A,T3,M) & M > 0 & next(M,T1,A,T4,P). 
next(M,cat(T1,T2),A,cat(T3,T2),N) :- next(M,T1,A,T3,N).
next(M,cat(T1,T2),A,T3,N) :- empty(T1) & next(M,T2,A,T3,N).

/****************************************************************/
/********** 				EMPTY PREDICATE				*********/
/****************************************************************/

empty(lambda).
empty(choice([T1,_])) :- empty(T1).
empty(choice([_,T1])) :- empty(T1).
empty(fork(T1,T2)) :- empty(T1) & empty(T2).
empty(cat(T1,T2)) :- empty(T1) & empty(T2).

/****************************************************************/
/********** 			TYPE_CHECK PREDICATE			*********/
/****************************************************************/

type_check(msg(S, R, P, C), NewState) :- 
   current_state(LastState, CId) &
   next(0, LastState, msg(S, R, P, C), NewState, 0).


/****************************************************************/
/********** 				CURRENT STATE				*********/
/****************************************************************/   
   
current_state(initial).
current_id(0).

/****************************************************************/
/********** 				INITIAL GOAL				*********/
/****************************************************************/

//!test(global_type(brokering_conv)).

/* Plans */

/* Monitor's plans */

/* @test[atomic]
+!test(global_type(Name)) : current_state(initial) & global_type(Name, G) 
   <- !move_to_state(G). */
   
@conversStartReq[atomic]
+start_conversation(Name)[source(S)]: 
current_id(Id) & NewId = Id+1 & global_type(Name, G) & CId = conv(S, Id)
   <-  -start_conversation(Name)[source(S)]; .print("\nMessage ", start_conversation(Name), " from source ", S, " causes the start of the conversation\n"); -current_id(Id); .print("\n1\n"); 
       +current_id(NewId); .print("\n2\n"); 
	   .send(S, tell, conversation_id(Name, CId)); .print("\n3\n"); 
	   +ongoing_conversation(Name, CId, S); .print("\n4\n"); 
	   +current_state(initial, CId); .print("\n5\n"); 
	   !move_to_state(G, CId); .print("\n5.5\n").   

@conversStopReq1[atomic]
+stop_conversation(CId)[source(S)]: 
ongoing_conversation(Name, CId, S) & current_state(LastState, CId) & empty(LastState)
   <-  -stop_conversation(CId)[source(S)]; 
       -ongoing_conversation(Name, CId, S);
	   +closed_conversation(Name, CId, S);
	   -current_state(LastState, CId);
       .print("\nMessage ", stop_conversation(CId), " from source ", S, " causes the succesful termination of the conversation\n").  	   

@conversStopReq2[atomic]
+stop_conversation(CId)[source(S1)]: 
ongoing_conversation(Name, CId, S2) & not(S1=S2)
   <-  .print("\nMessage ", stop_conversation(CId), " from source ", S1, " cannot stop a conversation started by ", S2, ": request ignored\n"). 

@conversStopReq3[atomic]
+stop_conversation(CId)[source(S)]: 
ongoing_conversation(Name, CId, S) & current_state(LastState, CId) & not(empty(LastState))
   <-  .print("\nMessage ", stop_conversation(CId), " from source ", S, " in state ", LastState, " cannot be accepted since the state is not empty: request ignored\n").   
   
@move2state[atomic]
+!move_to_state(NewState, CId) : current_state(LastState, CId) 
   <- -current_state(_, CId); .print("\n6\n"); 
      +current_state(NewState, CId); .print("\n7\n"). 

@successfulMove[atomic]
+!type_check_message(msg(S, R, P, m(C, CId))) : current_state(Current, CId) & ongoing_conversation(Name, CId, _) & type_check(msg(S, R, P, m(C, CId)), NewState)  
   <- !move_to_state(NewState, CId);
      .print("\nMessage ", msg(S, R, P, m(C, CId)), "\nleads from state ", Current, " to state ", NewState, "\n");
	  .wait(500);
      .send(S, tell, ok_check(msg(S, R, P, m(C, CId)))).

@failingMoveAndProtocol
+!type_check_message(msg(S, R, P, m(C, CId))) : current_state(Current, CId) 
                                        & ongoing_conversation(Name, CId, Source)
   <- .print("\n*** DYNAMIC TYPE-CHECKING ERROR ***\nMessage ", msg(S, R, P, m(C, CId)), " received within protocol ", Name, "\ncannot be accepted in the current state ", Current, "\n"); -ongoing_conversation(Name, CId, Source); -current_state(Current, CId); +failed_conversation(Name, CId, Source).
	   
@messageReceptionOK
+msg(S, R, P, m(C, CId))[source(S)]: .print("\n8: Received ", msg(S, R, P, m(C, CId))[source(S)], "\n") & ongoing_conversation(_Name, CId, _Source) & timeout(MS)
   <- -msg(S, R, P, m(C, CId))[source(S)]; .print("\nReceived ", msg(S, R, P, m(C, CId))[source(S)], "\n");
      !type_check_message(msg(S, R, P, m(C, CId)));
	  !check_progress.   
   
@messageReceptionKO1
+msg(S, R, P, m(C, CId)): failed_conversation(Name, CId, Source)
   <- .print("\n*** DYNAMIC TYPE-CHECKING ERROR ***\nMessage ", msg(S, R, P, m(C, CId)), " received but conversation ", CId, " started by ", Source, " within protocol ", Name, " previously failed\n").

@messageReceptionKO2
+msg(S, R, P, m(C, CId)): closed_conversation(Name, CId, Source)
   <- .print("\n*** DYNAMIC TYPE-CHECKING ERROR ***\nMessage ", msg(S, R, P, m(C, CId)), " received but conversation ", CId, " started by ", Source, " within protocol ", Name, " already closed successfully\n").
   
@messageReceptionKO3
+msg(S, R, P, m(C, CId)): not(failed_conversation(_, CId, _)) &               
                       not(ongoing_conversation(_, CId, _)) &
					   not(closed_conversation(_, CId, _))
   <- .print("\n*** DYNAMIC TYPE-CHECKING ERROR ***\nMessage ", msg(S, R, P, m(C, CId)), " received but conversation ", CId, " never started\n").
   
+!check_progress : timeout(MS) 
   <- .wait({+msg(S1, R1, P1, C1)}, MS, Event); 
	  !aux_check_progress(Event).
	  
+!aux_check_progress(Event) : timeout(MS) & Event < MS. 

+!aux_check_progress(Event) : timeout(MS) & Event >= MS  
   <- .print("\n*** WARNING ***\nNo progress for ", Event, " milliseconds\n"); 
   !check_progress. 
  
/* Plans for runtime type checking */

{include("./_session_tools_brokering_conv.asl")}   
