// Agent monitor in project infinite2.mas2j

/* Initial beliefs and rules */

timeout(4000). 

/****************************************************************/
/********** 				PROTOCOLS					*********/
/****************************************************************/

/* complex a^n b^n, involves alice_complex, bob_complex */
global_type(abn_complex, Forever) :- // a^n b^n repeated many times, with n varying each time (ICTCS 2012 paper), infinite
	Forever = cat(AB, Forever) & 
    AB = seq(sa(a, 0),
        choice([cat(AB, seq(sa(b, 0),lambda)), seq(sa(b, 0),lambda)])).			  
		
		
/****************************************************************/
/********** 		SENDING ACTION TYPES				*********/
/****************************************************************/

/* Types for the a^n b^n complex protocol */
has_type(msg(alice_complex, bob_complex, tell, a(X, T, N)), a) :- int(X) & int(T) & int(N).
has_type(msg(bob_complex, alice_complex, tell, b(X, N)), b) :- int(X) & int(N).

string(_).
int(_).
   

/****************************************************************/
/********** 				NEXT PREDICATE				*********/
/****************************************************************/

next(0,seq(sa(AType,N),T),AMessage,T,N) :- has_type(AMessage, AType).
next(M,seq(A,T1),A,T1,N) :- M > 0 & N = M - 1.
next(M,choice([T1,_]),A,T2,N) :- next(M,T1,A,T2,N). 
next(M,choice([_,T1]),A,T2,N) :- next(M,T1,A,T2,N).
next(N,fork(T1,T2),A,fork(T3,T2),M) :- next(N,T1,A,T3,M). 
next(N,fork(T1,T2),A,fork(T1,T3),M) :- next(N,T2,A,T3,M).
next(N,fork(T1,T2),A,fork(T3,T4),P) :- next(N,T1,A,T3,M) & M > 0 & next(M,T2,A,T4,P). 
next(N,fork(T1,T2),A,fork(T4,T3),P) :- next(N,T2,A,T3,M) & M > 0 & next(M,T1,A,T4,P). 
next(M,cat(T1,T2),A,cat(T3,T2),N) :- next(M,T1,A,T3,N).
next(M,cat(T1,T2),A,T3,N) :- empty(T1) & next(M,T2,A,T3,N).

/****************************************************************/
/********** 				EMPTY PREDICATE				*********/
/****************************************************************/

empty(lambda).
empty(choice([T1,_])) :- empty(T1).
empty(choice([_,T1])) :- empty(T1).
empty(fork(T1,T2)) :- empty(T1) & empty(T2).
empty(cat(T1,T2)) :- empty(T1) & empty(T2).

/****************************************************************/
/********** 			TYPE_CHECK PREDICATE			*********/
/****************************************************************/

type_check(msg(S, R, P, C), NewState) :- 
   current_state(LastState) &
   next(0, LastState, msg(S, R, P, C), NewState, 0).


/****************************************************************/
/********** 				CURRENT STATE				*********/
/****************************************************************/   
   
current_state(initial).

/****************************************************************/
/********** 				INITIAL GOAL				*********/
/****************************************************************/


!test(global_type(abn_complex)).

/* Plans */

/* Monitor's plans */

@test[atomic]
+!test(global_type(Name)) : current_state(initial) & global_type(Name, G) 
   <- !move_to_state(G).

@move2state[atomic]
+!move_to_state(NewState) : current_state(LastState) 
   <- -current_state(_);
      +current_state(NewState).	

@successfulMove[atomic]
+!type_check_message(msg(S, R, P, C)) : current_state(Current) & type_check(msg(S, R, P, C), NewState)  
   <- !move_to_state(NewState);
      .print("\nMessage ", msg(S, R, P, C), "\nleads from state ", Current, " to state ", NewState, "\n");
	  .wait(500);
      .send(S, tell, ok_check(msg(S, R, P, C))).

@failingMoveAndProtocol
+!type_check_message(msg(S, R, P, C)) : current_state(Current) 
                                        & global_type(Name, _)
   <- .print("\n*** DYNAMIC TYPE-CHECKING ERROR ***\nMessage ", msg(S, R, P, C), " received within protocol ", Name, "\ncannot be accepted in the current state ", Current, "\n");
       !move_to_state(initial).

@messageReceptionKO
+msg(S, R, P, C): current_state(initial)
   <- .print("\n*** DYNAMIC TYPE-CHECKING ERROR ***\nMessage ", msg(S, R, P, C), " received but no protocol is ongoing successfully\n").
   
@messageReceptionOK
+msg(S, R, P, C)[source(S)]: timeout(MS)
   <- -msg(S, R, P, C)[source(S)]; 
      !type_check_message(msg(S, R, P, C));
	  !check_progress.
	  
+!check_progress : timeout(MS) 
   <- .wait({+msg(S1, R1, P1, C1)}, MS, Event); 
	  !aux_check_progress(Event).
	  
+!aux_check_progress(Event) : timeout(MS) 
                                           & Event < MS. 

+!aux_check_progress(Event) : timeout(MS) 
                                           & current_state(Current) 
										   & Event >= MS  
   <- .print("\n*** WARNING ***\nNo progress for ", Event, " milliseconds in the current state ", Current, "\n"); 
   !check_progress. 
