// Agent monitor in project infinite2.mas2j

/* Initial beliefs and rules */

timeout(4000). 

/****************************************************************/
/********** 				PROTOCOLS					*********/
/****************************************************************/

/* Brokering protocol, involves b, c, s */
global_type(brokering, Glob) :- // Esop 2012 paper by Pierre-Malo Denielou and Nobuko Yoshida, page 2
   Merge = choice([Off,Fork]) &
   Off=
   seq(sa(offer, 0),
      seq(sa(counter, 0), Merge)) &
   Fork= fork(seq(sa(final, 0),lambda),
              seq(sa(result, 0),lambda)) &
   Glob = seq(sa(item, 0),Merge). 
   
/****************************************************************/
/********** 		SENDING ACTION TYPES				*********/
/****************************************************************/

/* Types for the brokering protocol */	  
has_type(msg(b, c, tell, offer(Item, Offer)), offer) :- string(Item) & int(Offer).
has_type(msg(c, b, tell, counter(Item, Offer)), counter) :- string(Item) & int(Offer).
has_type(msg(b, s, tell, final(Res, Client, Item, Offer)), final) :- string(Res) & string(Client) & string(Item) & int(Offer).
has_type(msg(b, c, tell, result(Res, Item, Offer)), result) :- string(Res) & string(Item) & int(Offer).
has_type(msg(s, b, tell, item(Client, Item)), item) :- string(Client) & string(Item).

string(_).
int(_).
   

/****************************************************************/
/********** 				NEXT PREDICATE				*********/
/****************************************************************/

next(0,seq(sa(AType,N),T),AMessage,T,N) :- has_type(AMessage, AType).
next(M,seq(A,T1),A,T1,N) :- M > 0 & N = M - 1.
next(M,choice([T1,_]),A,T2,N) :- next(M,T1,A,T2,N). 
next(M,choice([_,T1]),A,T2,N) :- next(M,T1,A,T2,N).
next(N,fork(T1,T2),A,fork(T3,T2),M) :- next(N,T1,A,T3,M). 
next(N,fork(T1,T2),A,fork(T1,T3),M) :- next(N,T2,A,T3,M).
next(N,fork(T1,T2),A,fork(T3,T4),P) :- next(N,T1,A,T3,M) & M > 0 & next(M,T2,A,T4,P). 
next(N,fork(T1,T2),A,fork(T4,T3),P) :- next(N,T2,A,T3,M) & M > 0 & next(M,T1,A,T4,P). 
next(M,cat(T1,T2),A,cat(T3,T2),N) :- next(M,T1,A,T3,N).
next(M,cat(T1,T2),A,T3,N) :- empty(T1) & next(M,T2,A,T3,N).

/****************************************************************/
/********** 				EMPTY PREDICATE				*********/
/****************************************************************/

empty(lambda).
empty(choice([T1,_])) :- empty(T1).
empty(choice([_,T1])) :- empty(T1).
empty(fork(T1,T2)) :- empty(T1) & empty(T2).
empty(cat(T1,T2)) :- empty(T1) & empty(T2).

/****************************************************************/
/********** 			TYPE_CHECK PREDICATE			*********/
/****************************************************************/

type_check(msg(S, R, P, C), NewState) :- 
   current_state(LastState) &
   next(0, LastState, msg(S, R, P, C), NewState, 0).


/****************************************************************/
/********** 				CURRENT STATE				*********/
/****************************************************************/   
   
current_state(initial).

/****************************************************************/
/********** 				INITIAL GOAL				*********/
/****************************************************************/

!test(global_type(brokering)).

/* Plans */

/* Monitor's plans */

@test[atomic]
+!test(global_type(Name)) : current_state(initial) & global_type(Name, G) 
   <- !move_to_state(G).

@move2state[atomic]
+!move_to_state(NewState) : current_state(LastState) 
   <- -current_state(_);
      +current_state(NewState).	

@successfulMove[atomic]
+!type_check_message(msg(S, R, P, C)) : current_state(Current) & type_check(msg(S, R, P, C), NewState)  
   <- !move_to_state(NewState);
      .print("\nMessage ", msg(S, R, P, C), "\nleads from state ", Current, " to state ", NewState, "\n");
	  .wait(500);
      .send(S, tell, ok_check(msg(S, R, P, C))).

@failingMoveAndProtocol
+!type_check_message(msg(S, R, P, C)) : current_state(Current) 
                                        & global_type(Name, _)
   <- .print("\n*** DYNAMIC TYPE-CHECKING ERROR ***\nMessage ", msg(S, R, P, C), " received within protocol ", Name, "\ncannot be accepted in the current state ", Current, "\n");
       !move_to_state(initial).

@messageReceptionKO
+msg(S, R, P, C): current_state(initial)
   <- .print("\n*** DYNAMIC TYPE-CHECKING ERROR ***\nMessage ", msg(S, R, P, C), " received but no protocol is ongoing successfully\n").
   
@messageReceptionOK
+msg(S, R, P, C)[source(S)]: timeout(MS)
   <- -msg(S, R, P, C)[source(S)]; 
      !type_check_message(msg(S, R, P, C));
	  !check_progress.
	  
+!check_progress : timeout(MS) 
   <- .wait({+msg(S1, R1, P1, C1)}, MS, Event); 
	  !aux_check_progress(Event).
	  
+!aux_check_progress(Event) : timeout(MS) 
                                           & Event < MS. 

+!aux_check_progress(Event) : timeout(MS) 
                                           & current_state(Current) 
										   & Event >= MS  
   <- .print("\n*** WARNING ***\nNo progress for ", Event, " milliseconds in the current state ", Current, "\n"); 
   !check_progress. 
